"""
A Python and command line tool for calculating coordination
networks of various types. It is primarily aimed at social media data, and the
examples and this is reflected in the names and examples used.

The input data for the preprocess command is a CSV file, with a header row,
containing fields in the following order:

- message_id: a unique id for the message. Messaged will be deduplicated using this
    field.
- user_id: a unique identifier for the user who posted the message.
- repost_id: for platforms that allow unedited broadcasting of other messages on the
    platform (such as retweets), this is the identifier of the original message.
    If this is an empty string this will be converted to null. If this field is
    populated, then co_link and co_message network computations will exclude this
    message, as a plain rebroadcast is always a duplicate.
- message: the text of the message.
- timestamp: the unix epoch timestamp of the message, in integer seconds.
- urls: a space delimited string containing urls. For Twitter URLs, it is better to
    use the expanded_url, not the default short_url Twitter automatically generates.
    If you use the short_url, co-link analysis will not be useful without running
    the resolver.
"""
import argparse
from coordination_network_toolkit.compute_networks import (compute_co_link_network,
    compute_co_reply_network, compute_co_retweet_parallel, compute_co_similar_tweet,
    compute_co_tweet_network)
from coordination_network_toolkit.preprocess import preprocess_csv_files, preprocess_twitter_json_files
from coordination_network_toolkit.output import write_output, output_node_csv
from coordination_network_toolkit.urls import resolve_all_urls
from coordination_network_toolkit.similarity import get_similarity_fn_from_min_size


_network_types = {
    'co_retweet': compute_co_retweet_parallel,
    'co_tweet': compute_co_tweet_network,
    'co_reply': compute_co_reply_network,
    'co_similar_tweet': compute_co_similar_tweet,
    'co_link': compute_co_link_network
}


def main():
    # TODO: how to best reference the docs?
    parser = argparse.ArgumentParser(
        prog="compute_networks.py",
        description=__doc__,
        epilog="For more information, see documentation at [docs url]",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument(
        "database",
        help="The path to the SQLite database holding the results. "
             "If it does not exist it will be created.",
    )

    subparsers = parser.add_subparsers(
        title="Commands",
        description="These are the available commands for coordination network processing. "
        "Each of the following sub-commands has further help information "
             "available by calling, for example, coordination_network_toolkit example.db preprocess --help",
        dest='command'
    )

    # preprocess subcommand
    preprocess_parser = subparsers.add_parser(
        'preprocess',
        help='Load data into a pre-processed format for network computation'
    )
    preprocess_parser.add_argument(
        "files",
        nargs="*",
        help="The files to be imported when running the preprocessing. Specify the "
             "format of the files with --format. All files must be the same format.",
    )
    preprocess_parser.add_argument(
        "--format",
        default="csv",
        choices=["csv", "twitter_json"],
        help="The format of the input files, defaulting to CSV. See documentation for a"
             " list of columns expected in CSV format."
    )

    # resolve_urls subcommand
    resolve_urls_parser = subparsers.add_parser(
        "resolve_urls",
        help="Resolve URLs in data with resolve_urls."
    )
    resolve_urls_parser.add_argument(
        "--max_redirects",
        type=int,
        default=1,
        help="The maximum number of redirects considered when resolving URLs",
    )

    # compute subcommand
    compute_parser = subparsers.add_parser(
        'compute',
        help="Compute network."
    )
    compute_parser.add_argument(
        'network_type',
        choices=_network_types.keys(),
        help="Which type of network to compute."
    )
    network_params_group = compute_parser.add_argument_group(
        'network_parameters',
        "Parameters for computing networks. Not all parameters apply to all types of"
        "network, please refer to the full documentation for details."
    )
    network_params_group.add_argument(
        "--time_window",
        type=int,
        default=1,
        help="The maximum number of seconds to consider when calculating coordination: "
        "if two events are recorded as further than this far apart in time they are "
        "not considered coordinated.",
    )
    network_params_group.add_argument(
        "--min_edge_weight",
        type=int,
        default=2,
        help="The minimum weight of an edge to be included in the output network. Edges "
        "are weighted by the number of tweets that are identified as coordinated between "
        "two users. A small mininum edge weight may include a very large number of "
        "spurious edges that indicate coincidence rather than coordination.",
    )
    network_params_group.add_argument(
        "--similarity_threshold",
        type=float,
        default=0.8,
        help="Only used for co-similarity networks. The similarity threshold for considering two tweets as similar "
        "documents. The default similarity function is Jaccard similarity and is "
        "normalised between 0 and 1.",
    )
    network_params_group.add_argument(
        "--min_document_size_similarity",
        type=float,
        default=1,
        help="Only used for co-similarity networks. The minimum document size (in tokens) to be considered when calculating "
        "document similarity. Documents smaller than this size are considered to have "
        "zero similarity with any other document.",
    )
    network_params_group.add_argument(
        "--resolved",
        action="store_true",
        help="Only used for co-link networks. When caculating a co-link network, use the resolved urls instead of the "
             "original URLs. Default is False. Requires that augment_data resolve_urls has "
             "been run beforehand, otherwise there will be no edges found.",
    )
    network_params_group.add_argument(
        "--n_cpus",
        type=int,
        default=2,
        help="The number of threads to use when calculating the co_retweet networks. "
             "Ignored for all other operations.",
    )
    network_params_group.add_argument(
        "--output_file",
        default=None,
        help="(optional) Immediately save the output to this file."
    )
    network_params_group.add_argument(
        "--output_format",
        choices=["csv", "graphml"],
        default="graphml",
        help="What format to export the network graph in. Default is graphml. "
        "Note that graphml output requires the whole graph to be loaded into "
        "memory before loading, and may not be appropriate for large graphs, or "
        "with low values for --min_edge_weight. Ignored if --output_file is not set.",
    )
    network_params_group.add_argument(
        "--include_symmetric_edges",
        help="Include the duplicate edges in the opposite direction. "
        "Default is False, so if there is an edge between A and B, there will be no "
        "edge recorded between B and A.",
        action="store_true",
    )
    network_params_group.add_argument(
        "--include_self_loops",
        help="Include loops (where users interact with their own posts.). Default is "
        "not to include loops.",
        action="store_true",
    )
    network_params_group.add_argument(
        "--n_messages",
        help="Number of messages to include in annotating a node's description.",
        type=int,
        default=10,
    )

    network_export_parser = subparsers.add_parser(
        'export_network',
        help='Export computed network edges to a file'
    )
    network_export_parser.add_argument(
        "output_file",
        help="Where to save the calculated output file."
    )
    network_export_parser.add_argument(
        'network_type',
        choices=_network_types.keys(),
        help='Which network to export. Network type must have previously been computed.'
    )
    network_export_parser.add_argument(
        "--output_format",
        choices=["csv", "graphml"],
        default="graphml",
        help="What format to export the network graph in. Default is graphml. "
        "Note that graphml output requires the whole graph to be loaded into "
        "memory before loading, and may not be appropriate for large graphs, or "
        "with low values for --min_edge_weight",
    )
    network_export_parser.add_argument(
        "--include_symmetric_edges",
        help="Include the duplicate edges in the opposite direction. "
        "Default is False, so if there is an edge between A and B, there will be no "
        "edge recorded between B and A.",
        action="store_true",
    )
    network_export_parser.add_argument(
        "--include_self_loops",
        help="Include loops (where users interact with their own posts.). Default is "
        "not to include loops.",
        action="store_true",
    )
    network_export_parser.add_argument(
        "--n_messages",
        help="Number of messages to include in annotating a node's description.",
        type=int,
        default=10,
    )

    # export_user_nodes subparser
    export_user_nodes_subparser = subparsers.add_parser(
        'export_user_nodes',
        help="Create an additional file listing user nodes and including their attributes."
    )
    export_user_nodes_subparser.add_argument(
        'output_file',
        help="The location to save the user node list to. The list will be saved as a"
             "CSV file."
    )


    # Process arguments
    args = parser.parse_args()

    if args.command == "preprocess":

        if not args.files:
            parser.error(
                "No files specified to preprocess."
            )

        if args.format == "csv":
            preprocess_csv_files(args.database, args.files)
        elif args.format == "twitter_json":
            preprocess_twitter_json_files(args.database, args.files)

    elif args.command == "resolve_urls":
        print(f"Resolving URLs on {args.database}")

        resolve_all_urls(args.database, args.max_redirects)

    elif args.command == "compute":
        network_calculation_status = (
            "Calculating a {args.network_type} network on {args.database} "
            "with the following settings:\n"
            "    time_window: {args.time_window} seconds\n"
            "    min_edge_weight: {args.min_edge_weight} co-ocurring messages\n"
            "    n_cpus: {args.n_cpus} processors\n"
            "    output_file: {args.output_file}"
        )

        if args.network_type == "co_retweet":
            print(network_calculation_status.format(args=args))
            compute_co_retweet_parallel(
                args.database,
                args.time_window,
                n_threads=args.n_cpus,
                min_edge_weight=args.min_edge_weight,
            )

        elif args.network_type == "co_tweet":
            print(network_calculation_status.format(args=args))
            compute_co_tweet_network(
                args.database, args.time_window, min_edge_weight=args.min_edge_weight
            )

        elif args.network_type == "co_reply":
            print(network_calculation_status.format(args=args))
            compute_co_reply_network(
                args.database, args.time_window, min_edge_weight=args.min_edge_weight
            )

        elif args.network_type == "co_similar_tweet":
            print(network_calculation_status.format(args=args))

            compute_co_similar_tweet(
                args.database,
                args.time_window,
                min_edge_weight=args.min_edge_weight,
                similarity_threshold=args.similarity_threshold,
                similarity_function=get_similarity_fn_from_min_size(args.min_document_size_similarity),
            )

        elif args.network_type == "co_link":
            print(network_calculation_status.format(args=args))
            print(f"    resolved: {args.resolved}")
            compute_co_link_network(
                args.database,
                args.time_window,
                min_edge_weight=args.min_edge_weight,
                resolved=args.resolved,
            )

        if args.output_file:
            print(args)
            write_output(
                args.database, args.network_type, args.output_file,
                output_type=args.output_format,
                symmetric=args.include_symmetric_edges,
                loops=args.include_self_loops,
                n_messages=args.n_messages
            )

    elif args.command == "export_network":
        write_output(
            args.database, args.network_type, args.output_file,
            output_type=args.output_format,
            symmetric=args.include_symmetric_edges,
            loops=args.include_self_loops,
            n_messages=args.n_messages
        )

    elif args.command == "export_user_nodes":
        print("Generating a user node annotation file")
        output_node_csv(args.database, args.output_file)


if __name__ == "__main__":
    main()



import csv
import random
import datetime as dt


header = [
    "message_id",
    "user_id",
    "username",
    "repost_id",
    "reply_id",
    "message",
    "timestamp",
    "urls"
]


words = [
    "apple",
    "pear",
    "banana",
    "pineapple",
    "mango",
    "cat",
    "dog",
    "squirrel",
    "reindeer",
    "wombat",
    "camel",
    "notebook",
    "pen",
    "pencil",
    "graph"
]


def generate_sample_data(file_path, num_messages=100, num_users=10, num_urls=10, example_domain="http://example.com/"):
    users = list()
    for i in range(0, num_users):
        users.append((i, "user_" + str(i)))

    urls = list()
    for i in range(0, num_urls):
        urls.append(example_domain + str(i))

    message_ids = list()
    with open(file_path, 'w') as fh:
        writer = csv.writer(fh)

        writer.writerow(header)

        for message_id in range(1, num_messages + 1):
            msg = [message_id]

            # user_id and username
            msg.extend(random.choice(users))

            # repost and reply ids
            # Areas for possible improvement:
            # - configurable probabilities
            # - possibility of values outside existing data
            # - should it be allowed to have both repost_id and reply_id?
            if len(message_ids) == 0:
                repost_id = None
                reply_id = None
            else:
                repost_id = random.choice(message_ids) if random.random() <= 0.3 else None
                reply_id = None
                if not repost_id and random.random() <= 0.3:
                    reply_id = random.choice(message_ids)
            msg.extend([repost_id, reply_id])

            # message
            # Areas for possible improvement:
            # - how to manage likelihood of repeat messages? Same as message_ids?
            message_text = " ".join(random.sample(words, random.randint(1, 4)))
            msg.append(message_text)

            # timestamp
            # They'll *all* be within the threshold like this
            msg.append(dt.datetime.now().timestamp())

            # urls
            # Areas for possible improvement:
            # - probably should be more likely that no URLs are included
            msg.append(random.sample(urls, random.randint(0, 2)))

            writer.writerow(msg)
            message_ids.append(message_id)


if __name__ == '__main__':
    generate_sample_data("sample_data.csv")



"""
Tests of the graph construction procedure.

"""
import os

from coordination_network_toolkit import preprocess, compute_networks, graph


def test_graph_construction_workflow(tmpdir):

    source_data = "tests/data/coretweet_test_input.csv"
    db = os.path.join(tmpdir, "test.db")

    preprocess.preprocess_csv_files(db, [source_data])

    settings_nodes = (
        # (Time window, min_edge_weight, loops in the output), expected nodes
        ((1, 1, False), 2),
        ((1, 1, True), 9),  # Self loops will always include a single retweet
        ((1, 2, False), 0),  # No edges over the cutoff weight
        ((1, 2, True), 5),  # 5 accounts with more than one retweet for the self loop
        # Note that in this example, uid10 retweets a message twice - this is
        # counted as a self loop so not included in the output nodes.
        ((60, 1, False), 6),  
        ((60, 1, True), 9),  # All of the nodes should be included with self loops
    )

    for n_threads in (1, 2, 4):

        for (time_window, min_edge_weight, self_loops), nodes in settings_nodes:
            compute_networks.compute_co_retweet_parallel(
                db, time_window, min_edge_weight=min_edge_weight, n_threads=n_threads
            )

            g = graph.load_networkx_graph(
                db, "co_retweet", symmetric=True, loops=self_loops, n_messages=10
            )

            print(time_window, min_edge_weight, self_loops)
            assert len(g.nodes) == nodes



import csv
import networkx as nx

from coordination_network_toolkit.graph import get_node_rows, get_edge_rows, load_networkx_graph


def output_node_csv(db_path, output_file, n_messages=10):
    """
    Annotate a node with a sample of their n_messages most recent posts.

    This can be imported into tools like Gephi to complement the edge lists generated by
    networks.

    """

    with open(output_file, "w") as f:
        writer = csv.writer(f, quoting=csv.QUOTE_ALL)
        writer.writerow(
            ["Id", "username"] + [f"message_{i}" for i in range(n_messages)]
        )

        for row in get_node_rows(db_path, n_messages=n_messages):
            writer.writerow(row)


def output_gephi_csv(db_path, command, output_file, symmetric=False, loops=False):

    with open(output_file, "w") as f:
        writer = csv.writer(f, quoting=csv.QUOTE_ALL)

        writer.writerow(["source", "target", "weight", "edge_type"])

        rows = get_edge_rows(db_path, command, symmetric=symmetric, loops=loops)

        for row in rows:
            writer.writerow(row)


def output_graphml(
    db_path, command, output_file, symmetric=False, loops=False, n_messages=10
):
    """
    Output a graphml file, representing the nodes and edges of the given table.

    """

    graph = load_networkx_graph(
        db_path, command, symmetric=symmetric, loops=loops, n_messages=n_messages
    )

    nx.write_graphml(graph, output_file)


def write_output(database_name, network_type, output_filename, output_type="csv", **kwargs):
    """Write the output, based on the given arguments."""
    if output_type == "graphml":
        output_fn = output_graphml
    elif output_type == "csv":
        output_fn = output_gephi_csv
    else:
        raise ValueError(f"{output_type} is not recognised as an available export format")

    output_fn(database_name, network_type, output_filename, **kwargs)



import urllib3
import requests
import sqlite3 as lite
from concurrent.futures import ThreadPoolExecutor, wait
import time
from logging import getLogger


logger = getLogger(__name__)


# Disable the warnings: we're going to try validating certs first, falling
# back to unverified second, setting a flag if it couldn't be verified.
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def resolve_url(
    url, timeout=(15.05, 15), from_header=None, max_redirects=5, verify_ssl=True
):
    """
    Follow redirects of a URL to the end of the chain.

    The final URL in the chain (regardless of the status code at that point) is the
    'resolved' URL.

    Parameters:

    url: the URL to follow.

    timeout: the timeout to consider before marking the connection as failed.

    from_header: an optional email address to add to the from header, so you can be a
        good netizen.

    max_redirects: the number of redirects to take before stopping.

    Returns:

        (url, resolved_url, final_status_code)

    In the case that the request times out, or the max number of redirects is
    reached, that will be recorded as a string error 'TIMEOUT' or
    'TOO_MANY_REDIRECTS'.

    An attempt will be made to verify the certificates of the redirects: if
    an SSL error is raised SSLError will be reported as the status.

    """

    # Create a requests session to set the max_redirects, and make sure to include a
    # from header in the requests so we are contactable if we're causing problems.
    session = requests.Session()

    if from_header:
        session.headers.update({"From": from_header})

    # The default is 30, we're reducing it a lot here, to avoid hammering servers in
    # the case of a redirect loop.
    session.max_redirects = max_redirects

    # If we cannot resolve the URL, it will be left at the original URL.
    resolved_url = url

    try:
        r = session.head(url, allow_redirects=True, timeout=timeout)
        status_code = r.status_code
    except requests.exceptions.SSLError as e:
        if verify_ssl:
            return resolve_url(
                url,
                timeout=timeout,
                from_header=from_header,
                max_redirects=max_redirects,
                verify_ssl=False,
            )
        else:
            status_code = "SSLError"
            r = e.response
    except requests.Timeout as e:
        status_code = "Timeout"
        r = e.response
    except requests.TooManyRedirects as e:
        status_code = "TooManyRedirects"
        r = e.response
    except requests.ConnectionError as e:
        status_code = "ConnectionError"
        r = e.response
    except Exception as e:
        status_code = "OtherError"
        r = None
        logger.exception(e)

    if r is not None:
        resolved_url = r.url

    return url, resolved_url, verify_ssl, status_code


def resolve_all_urls(db_path, max_redirects):
    """
    Resolve all unresolved URLs in the given database.

    Resolution is rate limited to 25 URLs per second.

    The database itself is used as the queue for managing what work needs to be done.

    :param max_redirects: The maximum number of redirects to allow before terminating
    resolution.

    """
    db = lite.connect(db_path)

    pool = ThreadPoolExecutor(50)
    # While it would be nice to not have to track the futures ourselves, the
    # alternative would be to either manage a separate queue of URLs to process
    # or just to hand over every single URL to the pool on startup.
    future_list = set()

    with db:
        to_resolve = [
            row[0]
            for row in db.execute(
                "select url from resolved_url where resolved_url is null",
            )
        ]

    n_urls_to_pause = 25

    n_to_resolve = len(to_resolve)
    n_resolved = 0

    print(f"Resolving {n_to_resolve} URLs")

    for url in to_resolve:

        future = pool.submit(resolve_url, url, max_redirects=max_redirects)
        future_list.add(future)
        n_urls_to_pause -= 1

        # This is the rate limiting step - we only add to the queue 125 URLs
        # at a time, then we pause to prevent too much concurrency.
        # This means the maximum rate at which we can resolve urls is 25/second.
        if n_urls_to_pause == 0:
            n_urls_to_pause = 25
            time.sleep(1)

        # When the inflight is too big, move onto the next one
        if len(future_list) >= 1000:
            completed, not_completed = wait(future_list, timeout=0)
            future_list = not_completed

            with db:
                for f in completed:
                    row = f.result()
                    db.execute("replace into resolved_url values(?, ?, ?, ?)", row)

                    n_resolved += 1
            print(f"Resolved {n_resolved} of {n_to_resolve} URLs")

    else:
        # Wait for everything to be done, before starting again
        completed, _ = wait(future_list)

        with db:
            for f in completed:
                row = f.result()
                db.execute("replace into resolved_url values(?, ?, ?, ?)", row)

    print("Constructing new resolved url tables")

    with db:
        db.executescript(
            """
            drop table if exists resolved_message_url;

            create table resolved_message_url(
                message_id references edge(message_id),
                resolved_url,
                timestamp,
                user_id,
                primary key (message_id, resolved_url)
            );

            insert or ignore into resolved_message_url
            select 
                message_id,
                resolved_url.resolved_url,
                timestamp,
                user_id
            from message_url
            inner join resolved_url using(url)
            """
        )



import json
import zlib
import csv

from coordination_network_toolkit.database import initialise_db

from typing import Iterable, List


def preprocess_csv_files(db_path: str, input_filenames: List[str]):
    for message_file in input_filenames:
        # Skip the header
        print(f"Begin preprocessing {message_file} into {db_path}")

        with open(message_file, "r") as messages:
            reader = csv.reader(messages)
            # Skip header
            next(reader)
            preprocess_data(db_path, reader)

        print(f"Done preprocessing {message_file} into {db_path}")


def preprocess_data(db_path: str, messages: Iterable):
    """

    Add messages to the dataset from the specified CSV files..

    Messages should be an iterator of messages with the content for each message
    in the following order:

    - message_id: the unique identifier of the message on the platform
    - user_id: the unique identifier of the user on the platform
    - username: the text of the username (only used for display)
    - repost_id: if the message is a verbatim report of another message (such as a retweet 
        or reblog), this is the identifier of that other message. Empty strings will be 
        converted to null
    - reply_id: if the message is in reply to another message, the identifier for that other
        message. Empty strings will be converted to null.
    - message: the text of the message.
    - timestamp: A timestamp in seconds for the message. The absolute offset does not matter,
        but it needs to be consistent across all rows
    - urls: A space delimited string containing all of the URLs in the message

    """

    db = initialise_db(db_path)

    try:
        db.execute("begin")
        processed = (
            (
                message_id,
                user_id,
                username,
                repost_id or None,
                reply_id or None,
                message,
                len(message),
                zlib.adler32(message.encode("utf8")),
                # This will be populated only when similarity calculations are necessary
                None,
                float(timestamp),
                urls.split(" ") if urls else [],
            )
            for message_id, user_id, username, repost_id, reply_id, message, timestamp, urls in messages
        )

        for row in processed:
            db.execute(
                "insert or ignore into edge values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                row[:-1],
            )

            message_id, user_id = row[:2]
            timestamp = row[-2]

            # Ignore url shared in reposts
            if not row[3]:
                for url in row[-1]:
                    db.execute(
                        "insert or ignore into message_url values(?, ?, ?, ?)",
                        (message_id, url, timestamp, user_id),
                    )

        db.execute("commit")
    finally:
        db.close()


def preprocess_twitter_json_files(db_path: str, input_filenames: List[str]):
    for message_file in input_filenames:
        # Skip the header
        print(f"Begin preprocessing {message_file} into {db_path}")
        with open(message_file, "r") as tweets:
            preprocess_twitter_json_data(db_path, tweets)

        print(f"Done preprocessing {message_file} into {db_path}")


def preprocess_twitter_json_data(db_path: str, tweets: Iterable[str]):
    """

    Add messages to the dataset from the specified tweets in Twitter JSON format.

    Tweets must be in Twitter JSON format as collected from the v1.1 JSON API.

    """

    db = initialise_db(db_path)

    try:
        db.execute("begin")

        for raw_tweet in tweets:

            tweet = json.loads(raw_tweet)

            # Try grabbing the full_text field from the extended format, otherwise
            # check if there's a extended_tweet object.
            # print(sorted(tweet.keys()))
            if "full_text" in tweet:
                tweet_text = tweet["full_text"]
            elif "extended_tweet" in tweet:
                tweet_text = tweet["extended_tweet"]["full_text"]
            else:
                tweet_text = tweet["text"]

            retweet = tweet.get("retweeted_status", {})

            row = (
                tweet["id_str"],
                tweet["user"]["id_str"],
                tweet["user"]["screen_name"],
                retweet.get("id_str"),
                tweet.get("in_reply_to_status_id_str", None),
                tweet_text,
                len(tweet_text),
                zlib.adler32(tweet_text.encode("utf8")),
                # This will be populated only when similarity calculations are necessary
                None,
                # Twitter epoch in seconds
                (int(tweet["id"]) >> 22) / 1000,
            )

            db.execute(
                "insert or ignore into edge values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", row,
            )

            # If it's a retweet, don't consider any of the urls as candidates
            if not row[3]:
                if retweet:
                    url_entities = retweet.get("extended_tweet", retweet)["entities"][
                        "urls"
                    ]
                else:
                    url_entities = tweet.get("extended_tweet", tweet)["entities"]["urls"]

                message_id, user_id = row[:2]
                timestamp = row[-1]

                urls = [u["expanded_url"] for u in url_entities]
                # Ignore urls shared in reposts
                for url in urls:
                    db.execute(
                        "insert or ignore into message_url values(?, ?, ?, ?)",
                        (message_id, url, timestamp, user_id),
                    )

        db.execute("commit")
    finally:
        db.close()



import sqlite3 as lite


# Provides the mapping from the CLI command to the table in the database.
COMMAND_TABLE_MAPPING = {
    "co_retweet": "co_retweet_network",
    "co_tweet": "co_tweet_network",
    "co_reply": "co_reply_network",
    "co_link": "co_link_network",
    "co_similar_tweet": "co_similar_tweet_network",
}


def initialise_db(db_path: str):
    """Initialise the database, ensuring the correct schema is in place."""

    db = lite.connect(db_path, isolation_level=None)

    db.executescript(
        """
        pragma journal_mode=WAL;
        pragma synchronous=normal;

        create table if not exists edge (
            message_id primary key,
            user_id not null,
            username text,
            repost_id,
            reply_id,
            message text,
            message_length integer,
            message_hash blob,
            token_set text,
            timestamp integer
        );

        create table if not exists message_url(
            message_id references edge(message_id),
            url,
            timestamp,
            user_id,
            primary key (message_id, url)
        );

        create table if not exists resolved_url(
            url primary key,
            resolved_url,
            ssl_verified,
            resolved_status
        );

        create trigger if not exists url_to_resolve after insert on message_url 
            begin
                insert or ignore into resolved_url(url) values(new.url);
            end;

        """
    )

    return db



from concurrent.futures import ThreadPoolExecutor, wait, FIRST_COMPLETED
import math
import sqlite3 as lite
import threading
from typing import Callable

from coordination_network_toolkit.similarity import similarity, tokenize


local = threading.local()


network_queries = {
    "co_tweet": """
        -- This query will create the table representing the edges of the network.
        create table co_tweet_network as
        select
            e_1.user_id as user_1,
            e_2.user_id as user_2,
            count(*) as weight
        from edge e_1
        inner join edge e_2
            on (e_1.message_length, e_1.message_hash, e_1.message) =
               (e_2.message_length, e_2.message_hash, e_2.message)
            and e_2.timestamp between e_1.timestamp - ?1 and e_1.timestamp + ?1
            and e_1.repost_id is null
            and e_2.repost_id is null
        group by e_1.user_id, e_2.user_id
        having weight >= ?2
        """,
    "co_reply": """
        -- This query will create the table representing the edges of the network.
        create table co_reply_network as
        select
            e_1.user_id as user_1,
            e_2.user_id as user_2,
            count(*) as weight
        from edge e_1
        inner join edge e_2
            on e_1.reply_id = e_2.reply_id
            and e_2.timestamp between e_1.timestamp - ?1 and e_1.timestamp + ?1
            and e_1.repost_id is null
            and e_2.repost_id is null
        group by e_1.user_id, e_2.user_id
        having weight >= ?2
        """,
    "co_similar_tweet": """
        -- This query will create the table representing the edges of the network.
        create table co_similar_tweet_network as
        select
            e_1.user_id as user_1,
            e_2.user_id as user_2,
            count(*) as weight
        from edge e_1 indexed by user_time
        inner join edge e_2
            -- Length filtering of the messages
            on e_2.timestamp between e_1.timestamp - ?1 and e_1.timestamp + ?1
        -- Note that this will only work where the Python similarity function has been
        -- registered on the connection - this is not a SQLite native function.
        where 
            e_1.repost_id is null
            and e_2.repost_id is null
            and e_1.token_set is not null
            and e_2.token_set is not null
            and similarity(e_1.token_set, e_2.token_set) >= ?3
        group by e_1.user_id, e_2.user_id
        having weight >= ?2
        """,
    "co_link": """
        -- This query will create the table representing the edges of the network.
        create table co_link_network as
        select
            e_1.user_id as user_1,
            e_2.user_id as user_2,
            count(*) as weight
        from message_url e_1
        inner join message_url e_2
            on e_1.url = e_2.url
            and e_2.timestamp between e_1.timestamp - ?1 and e_1.timestamp + ?1
        group by e_1.user_id, e_2.user_id
        having weight >= ?2
        """,
    "co_link_resolved": """
        -- This query will create the table representing the edges of the network.
        create table co_link_network as
        select
            e_1.user_id as user_1,
            e_2.user_id as user_2,
            count(*) as weight
        from resolved_message_url e_1
        inner join resolved_message_url e_2
            on e_1.resolved_url = e_2.resolved_url
            -- Keep any row where the retweets are by different users and within n
            -- seconds of each other.
            and e_2.timestamp between e_1.timestamp - ?1 and e_1.timestamp + ?1
        group by e_1.user_id, e_2.user_id
        having weight >= ?2
        """,
}


def compute_co_tweet_network(db_path, time_window, min_edge_weight=1):
    """

    """
    db = lite.connect(db_path)

    with db:
        print("Ensuring the necessary index exists")
        db.execute(
            """
            create index if not exists message_content on edge(
                message_length, message_hash, timestamp
            ) where repost_id is null
            """
        )
        db.execute("drop table if exists co_tweet_network")
        db.execute(
            network_queries["co_tweet"], [time_window, min_edge_weight],
        )


def compute_co_reply_network(db_path, time_window, min_edge_weight=1):
    """

    """
    db = lite.connect(db_path)

    with db:
        print("Ensuring the necessary index exists")
        db.execute("create index if not exists user_time on edge(user_id, timestamp)")
        db.execute(
            """
            create index if not exists replies on edge(reply_id, timestamp) 
            where repost_id is null
            """
        )
        db.execute("drop table if exists co_reply_network")
        db.execute(
            network_queries["co_reply"], [time_window, min_edge_weight],
        )


def compute_co_link_network(db_path, time_window, min_edge_weight=1, resolved=False):
    """

    """
    db = lite.connect(db_path)

    print("Ensuring the necessary index exists")

    with db:

        db.execute("drop table if exists co_link_network")

        if resolved:

            db.execute(
                """
                create index if not exists resolved_url_message on resolved_message_url(
                    resolved_url, timestamp
                )
                """
            )
            db.execute(
                network_queries["co_link_resolved"], [time_window, min_edge_weight],
            )

        else:

            db.execute(
                """
                create index if not exists url_message on message_url(
                    url, timestamp
                )
                """
            )

            db.execute(
                network_queries["co_link"], [time_window, min_edge_weight],
            )


def compute_co_similar_tweet(
    db_path,
    time_window,
    n_threads=4,
    similarity_threshold=0.9,
    min_edge_weight=1,
    similarity_function: Callable = similarity,
):
    """  

    Create a network where tweets with a certain similarity are counted as coordinated
    edges.

    An arbitrary text similarity function can be applied. It must take the string 
    representation of the two messages to compared. The default is the Jaccard
    similarity, with a normalised similarity between 0 and 1.

    Currently the n_threads argument is not used.
    
    """
    db = lite.connect(db_path, isolation_level=None)
    db.create_function("similarity", 2, similarity_function)
    db.create_function("tokenize", 1, tokenize)

    db.executescript(
        """
        create index if not exists user_time on edge(user_id, timestamp);
        create index if not exists to_tokenize on edge(message_id)
            where repost_id is null and token_set is null;
        create index if not exists timestamp on edge(timestamp);
        drop table if exists co_similar_tweet_network;
        """
    )

    db.execute("begin")

    # Tokenize text
    print("Tokenizing messages")
    db.execute(
        """
        update edge set token_set = tokenize(message)
        where repost_id is null and token_set is null
        """
    )

    db.execute(
        network_queries["co_similar_tweet"],
        [time_window, min_edge_weight, similarity_threshold],
    )

    db.execute("commit")


def compute_co_retweet_parallel(db_path, time_window, n_threads=4, min_edge_weight=1):

    lock = threading.Lock()

    def calculate_user_edges(user_ids, time_window):
        try:
            db = local.db
        except AttributeError:
            db = lite.connect(db_path)
            db.executescript(
                """
                pragma synchronous=normal;
                create temporary table local_network (
                    user_id_1,
                    user_id_2,
                    weight
                );
                create temporary table user_id (
                    user_id primary key
                );
                """
            )
            local.db = db

        with db:
            db.execute("delete from user_id")
            db.execute("delete from local_network")
            db.executemany(
                "insert into user_id values(?)", ((user_id,) for user_id in user_ids)
            )

            db.execute(
                f"""
                insert into local_network
                    select
                        e_1.user_id as user_1,
                        e_2.user_id as user_2,
                        count(*) as weight
                        -- TODO: Add additional summary statistics
                    from edge e_1
                    inner join edge e_2
                        on e_1.repost_id = e_2.repost_id
                        and e_2.timestamp between e_1.timestamp - ?1 
                            and e_1.timestamp + ?1
                        and user_1 in (select user_id from user_id)
                    group by e_1.user_id, e_2.user_id
                    having count(*) >= ?
                """,
                [time_window, min_edge_weight],
            )

        with lock, db:
            db.execute("insert into co_retweet_network select * from local_network")

    db = lite.connect(db_path)

    print("Ensure the indexes exist to drive the join.")
    db.executescript(
        """
        create index if not exists user_time on edge(user_id, timestamp);
        create index if not exists repost_time on edge(
            repost_id, timestamp
        ) where repost_id is not null;
        drop table if exists co_retweet_network;
        create table co_retweet_network (
            user_1,
            user_2,
            weight,
            primary key (user_1, user_2)
        ) without rowid;
        """
    )

    pool = ThreadPoolExecutor(n_threads)

    print("Calculating the co-retweet network")
    waiting = set()
    count = 0
    completed = 0
    submitted = 0

    user_ids = []

    batch_size = 1000
    n_batches = math.ceil(
        list(db.execute("select count(distinct user_id) from edge"))[0][0] / batch_size
    )

    for (user_id,) in db.execute("select distinct user_id from edge"):

        user_ids.append(user_id)

        if len(user_ids) == batch_size:
            waiting.add(pool.submit(calculate_user_edges, user_ids, time_window))

            submitted += 1

            user_ids = []

            if len(waiting) >= 20:
                done, waiting = wait(waiting, return_when=FIRST_COMPLETED)

                for d in done:
                    d.result()
                    completed += 1

            if not (submitted % 100):
                print(f"Completed {completed} / {n_batches}")

    else:
        print("Waiting for final batch.")
        waiting.add(pool.submit(calculate_user_edges, user_ids, time_window))
        wait(waiting)

    db.close()

    return completed




import pytest

from coordination_network_toolkit.similarity import tokenize


# todo: does order matter?

tokens_tests = [
    ("hello", "hello"),
    ("The quick brown fox jumped over the lazy dog.", ". brown dog fox jumped lazy over quick the"),
    ("", "")
]


@pytest.mark.parametrize(["text", "tokens"], tokens_tests)
def test_tokenize(text, tokens):
    result = tokenize(text)
    assert result == tokens



import regex
from typing import Callable, Pattern


word_tokenizer = regex.compile(
    # Note this handles 'quoted' words a little weirdly: 'orange' is tokenised
    # as ["orange", "'"] I'd prefer to tokenise this as ["'", "orange", "'"]
    # but the regex library behaves weirdly. So for now phrase search for
    # quoted strings won't work.
    r"\b\p{Word_Break=WSegSpace}*'?",
    flags=regex.WORD | regex.UNICODE | regex.V1,
)


def tokenize(text: str, tokenizer: Pattern = word_tokenizer) -> str:
    words = sorted(set(t for t in tokenizer.split(text.lower()) if t))
    tokenized = " ".join(words)
    return tokenized


def similarity(tokens_1, tokens_2):
    set_1 = set(tokens_1.split())
    set_2 = set(tokens_2.split())
    return len(set_1 & set_2) / len(set_1 | set_2)


class MinDocSizeSimilarity:
    """
    A callable class for document similarity that discards short documents.

    This is designed to avoid considering extremely short documents (such as a tweet
    containing only a single mention and hashtag as similar in any way.

    Note that this is a callable class rather than a function to make future parallel
    processing easier.
    """

    def __init__(self, min_tokens=5):
        self.min_tokens = min_tokens

    def __call__(self, tokens_1, tokens_2) -> float:

        set_1 = set(tokens_1.split())
        if len(set_1) < self.min_tokens:
            return 0

        set_2 = set(tokens_2.split())
        if len(set_2) < self.min_tokens:
            return 0

        return len(set_1 & set_2) / len(set_1 | set_2)


def get_similarity_fn_from_min_size(min_document_size_similarity) -> Callable:
    if min_document_size_similarity > 1:
        return MinDocSizeSimilarity(min_document_size_similarity)
    else:
        return similarity



from coordination_network_toolkit.__main__ import main
import coordination_network_toolkit.compute_networks
import coordination_network_toolkit.database
import coordination_network_toolkit.graph
import coordination_network_toolkit.output
import coordination_network_toolkit.preprocess
import coordination_network_toolkit.similarity
import coordination_network_toolkit.urls



import sqlite3 as lite
import networkx as nx

from coordination_network_toolkit.database import COMMAND_TABLE_MAPPING


def get_edge_rows(db_path, command, symmetric=False, loops=False):
    """
    Return an iterator over the rows of the source table in the given database file.

    """

    if command not in COMMAND_TABLE_MAPPING:
        raise ValueError(f"No known table for command {command}.")

    table = COMMAND_TABLE_MAPPING[command]

    if symmetric and loops:
        query = f"select *, '{command}' from {table}"
    elif symmetric:
        query = f"select *, '{command}' from {table} where user_1 != user_2"
    elif loops:
        query = f"select *, '{command}' from {table} where user_2 >= user_1"
    else:
        query = f"select *, '{command}' from {table} where user_2 > user_1"

    db = lite.connect(db_path)

    rows = db.execute(query)

    return rows


def get_node_rows(db_path, n_messages=10):
    """Get an iterator of rows in the node table."""
    db = lite.connect(db_path)

    db.execute("create index if not exists user_time on edge(user_id, timestamp)")

    user_ids = (row[0] for row in db.execute("select distinct user_id from edge"))

    for user_id in user_ids:

        username = db.execute(
            "select max(username) from edge where user_id = ?", [user_id]
        ).fetchone()[0]

        tweets = [
            row[0]
            for row in db.execute(
                """
                select message 
                from edge
                where user_id = ?
                order by timestamp desc
                limit ? 
                """,
                [user_id, n_messages],
            )
        ]

        tweets.extend([""] * (n_messages - len(tweets)))

        yield [user_id, username] + tweets


def load_networkx_graph(db_path, command, symmetric=False, loops=False, n_messages=10):
    """Return a networkx graph object representing the given source table."""

    g = nx.Graph()

    edges = get_edge_rows(db_path, command, symmetric=symmetric, loops=loops)

    # Add the edges
    for user_1, user_2, weight, edge_type in edges:
        g.add_edge(user_1, user_2, weight=weight, edge_type=edge_type)

    # Add the node annotations
    nodes = get_node_rows(db_path, n_messages=n_messages)
    for row in nodes:
        user_id = row[0]

        # Only add the node annotations if the node is present from an edge
        if user_id in g.nodes:
            attrs = {"username": row[1]}
            for i, message in enumerate(row[2:]):
                attrs[f"message_{i}"] = message
            g.add_node(user_id, **attrs)

    return g



import setuptools

install_requires = [
    "urllib3",
    "regex==2019.11.1",
    "requests",
    "networkx>=2.5"
]

extras_require = {
    "development": [
        "nox",
        "pytest",
        "pip-tools"
    ]
}

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="coordination_network_toolkit",
    author="Timothy Graham (Digital Media Research Centre, QUT) and QUT Digital Observatory",
    author_email="digitalobservatory@qut.edu.au",
    description="Tools for computing networks of coordinated behaviour on social media",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/QUT-Digital-Observatory/coordination-network-toolkit",
    license='MIT',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Sociology"
    ],
    keywords='social_science social_media_analysis similarity_networks',
    python_requires='>=3.6',
    setup_requires=['setuptools_scm'],
    use_scm_version=True,
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points={
        'console_scripts': [
            'compute_networks=coordination_network_toolkit.__main__:main'
        ]
    }
)



import pytest

from coordination_network_toolkit.similarity import MinDocSizeSimilarity, similarity


@pytest.mark.parametrize(
    ["text_pair", "result"],
    [
        (("hello world", "hello world"), 1),
        (("hello world", "world hello"), 1),
    ]
)
def test_similarity(text_pair, result):
    assert similarity(*text_pair) == result
